# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from collections import namedtuple

Example = namedtuple('Example', ['problem', 'plan'])

def get_example_problems():
    problems = {}

    # robot
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    battery_charge = Fluent('battery_charge', RealType(0, 100))
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10))
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot')
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))])
    robot = Example(problem=problem, plan=plan)
    problems['robot'] = robot

    #robot fluent of user_type
    Location = UserType('Location')
    Robot = UserType('Robot')
    is_at = Fluent('is_at', Location, robot=Robot)
    move = InstantaneousAction('move', robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter('robot')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(Equals(is_at(robot), l_from))
    move.add_precondition(Not(Equals(is_at(robot), l_to)))
    move.add_effect(is_at(robot), l_to)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    r1 = Object('r1', Robot)
    r2 = Object('r2', Robot)
    problem = Problem('robot_fluent_of_user_type')
    problem.add_fluent(is_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(r1)
    problem.add_object(r2)
    problem.set_initial_value(is_at(r1), l2)
    problem.set_initial_value(is_at(r2), l1)
    problem.add_goal(Equals(is_at(r1), l1))
    problem.add_goal(Equals(is_at(r2), l2))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(r1), ObjectExp(l2), ObjectExp(l1))),
                                   up.plan.ActionInstance(move, (ObjectExp(r2), ObjectExp(l1), ObjectExp(l2)))])
    robot_fluent_of_user_type = Example(problem=problem, plan=plan)
    problems['robot_fluent_of_user_type'] = robot_fluent_of_user_type

    # robot no negative preconditions
    Location = UserType('location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(robot_at(l_from))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot')
    problem.add_fluent(robot_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.add_goal(robot_at(l2))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))])
    robot_no_negative_preconditions = Example(problem=problem, plan=plan)
    problems['robot_no_negative_preconditions'] = robot_no_negative_preconditions

    # robot decrease
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    battery_charge = Fluent('battery_charge', RealType(0, 100))
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_decrease_effect(battery_charge, 10)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_decrease')
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))])
    robot_decrease = Example(problem=problem, plan=plan)
    problems['robot_decrease'] = robot_decrease

    # robot_loader
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    cargo_at = Fluent('cargo_at', BoolType(), position=Location)
    cargo_mounted = Fluent('cargo_mounted')
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    load = InstantaneousAction('load',loc=Location)
    loc = load.parameter('loc')
    load.add_precondition(cargo_at(loc))
    load.add_precondition(robot_at(loc))
    load.add_precondition(Not(cargo_mounted))
    load.add_effect(cargo_at(loc), False)
    load.add_effect(cargo_mounted, True)
    unload = InstantaneousAction('unload', loc=Location)
    loc = unload.parameter('loc')
    unload.add_precondition(Not(cargo_at(loc)))
    unload.add_precondition(robot_at(loc))
    unload.add_precondition(cargo_mounted)
    unload.add_effect(cargo_at(loc), True)
    unload.add_effect(cargo_mounted, False)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_loader')
    problem.add_fluent(robot_at)
    problem.add_fluent(cargo_at)
    problem.add_fluent(cargo_mounted)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(cargo_at(l1), False)
    problem.set_initial_value(cargo_at(l2), True)
    problem.set_initial_value(cargo_mounted, False)
    problem.add_goal(cargo_at(l1))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2))),
                                   up.plan.ActionInstance(load, (ObjectExp(l2), )),
                                   up.plan.ActionInstance(move, (ObjectExp(l2), ObjectExp(l1))),
                                   up.plan.ActionInstance(unload, (ObjectExp(l1), ))])
    robot_loader = Example(problem=problem, plan=plan)
    problems['robot_loader'] = robot_loader

    # robot_loader_mod
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    cargo_at = Fluent('cargo_at', BoolType(), position=Location)
    is_same_location = Fluent('is_same_location', BoolType(), p1=Location, p2=Location)
    cargo_mounted = Fluent('cargo_mounted')
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_precondition(Not(is_same_location(l_from, l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    load = InstantaneousAction('load',loc=Location)
    loc = load.parameter('loc')
    load.add_precondition(cargo_at(loc))
    load.add_precondition(robot_at(loc))
    load.add_precondition(Not(cargo_mounted))
    load.add_effect(cargo_at(loc), False)
    load.add_effect(cargo_mounted, True)
    unload = InstantaneousAction('unload', loc=Location)
    loc = unload.parameter('loc')
    unload.add_precondition(Not(cargo_at(loc)))
    unload.add_precondition(robot_at(loc))
    unload.add_precondition(cargo_mounted)
    unload.add_effect(cargo_at(loc), True)
    unload.add_effect(cargo_mounted, False)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_loader_mod')
    problem.add_fluent(robot_at, default_initial_value=False)
    problem.add_fluent(cargo_at, default_initial_value=False)
    problem.add_fluent(cargo_mounted, default_initial_value=False)
    problem.add_fluent(is_same_location, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(cargo_at(l2), True)
    for o in problem.objects(Location):
        problem.set_initial_value(is_same_location(o, o), True)
    problem.add_goal(cargo_at(l1))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2))),
                                   up.plan.ActionInstance(load, (ObjectExp(l2), )),
                                   up.plan.ActionInstance(move, (ObjectExp(l2), ObjectExp(l1))),
                                   up.plan.ActionInstance(unload, (ObjectExp(l1), ))])
    robot_loader_mod = Example(problem=problem, plan=plan)
    problems['robot_loader_mod'] = robot_loader_mod

    # robot_loader_adv
    Robot = UserType('Robot')
    Container = UserType('Container')
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), robot=Robot, position=Location)
    cargo_at = Fluent('cargo_at', BoolType(), cargo=Container, position=Location)
    cargo_mounted = Fluent('cargo_mounted', BoolType(), cargo=Container, robot=Robot)
    move = InstantaneousAction('move', l_from=Location, l_to=Location, r=Robot)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    r = move.parameter('r')
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(r, l_from))
    move.add_precondition(Not(robot_at(r, l_to)))
    move.add_effect(robot_at(r, l_from), False)
    move.add_effect(robot_at(r, l_to), True)
    load = InstantaneousAction('load', loc=Location, r=Robot, c=Container)
    loc = load.parameter('loc')
    r = load.parameter('r')
    c = load.parameter('c')
    load.add_precondition(cargo_at(c, loc))
    load.add_precondition(robot_at(r, loc))
    load.add_precondition(Not(cargo_mounted(c, r)))
    load.add_effect(cargo_at(c, loc), False)
    load.add_effect(cargo_mounted(c,r), True)
    unload = InstantaneousAction('unload', loc=Location, r=Robot, c=Container)
    loc = unload.parameter('loc')
    r = unload.parameter('r')
    c = unload.parameter('c')
    unload.add_precondition(Not(cargo_at(c, loc)))
    unload.add_precondition(robot_at(r, loc))
    unload.add_precondition(cargo_mounted(c,r))
    unload.add_effect(cargo_at(c, loc), True)
    unload.add_effect(cargo_mounted(c,r), False)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    l3 = Object('l3', Location)
    r1 = Object('r1', Robot)
    c1 = Object('c1', Container)
    problem = Problem('robot_loader_adv')
    problem.add_fluent(robot_at)
    problem.add_fluent(cargo_at)
    problem.add_fluent(cargo_mounted)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(r1)
    problem.add_object(c1)
    problem.set_initial_value(robot_at(r1,l1), True)
    problem.set_initial_value(robot_at(r1,l2), False)
    problem.set_initial_value(robot_at(r1,l3), False)
    problem.set_initial_value(cargo_at(c1,l1), False)
    problem.set_initial_value(cargo_at(c1,l2), True)
    problem.set_initial_value(cargo_at(c1,l3), False)
    problem.set_initial_value(cargo_mounted(c1,r1), False)
    problem.add_goal(cargo_at(c1,l3))
    problem.add_goal(robot_at(r1,l1))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2), ObjectExp(r1))),
                                   up.plan.ActionInstance(load, (ObjectExp(l2), ObjectExp(r1), ObjectExp(c1))),
                                   up.plan.ActionInstance(move, (ObjectExp(l2), ObjectExp(l3), ObjectExp(r1))),
                                   up.plan.ActionInstance(unload, (ObjectExp(l3), ObjectExp(r1), ObjectExp(c1))),
                                   up.plan.ActionInstance(move, (ObjectExp(l3), ObjectExp(l1), ObjectExp(r1)))])
    robot_loader_adv = Example(problem=problem, plan=plan)
    problems['robot_loader_adv'] = robot_loader_adv

    #robot locations connected
    Location = UserType('Location')
    Robot = UserType('Robot')
    is_at = Fluent('is_at', BoolType(), position=Location, robot=Robot)
    battery_charge = Fluent('battery_charge', RealType(0, 100), robot=Robot)
    is_connected = Fluent('is_connected', BoolType(), location_1=Location, location_2=Location)
    move = InstantaneousAction('move', robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter('robot')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge(robot), 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move.add_decrease_effect(battery_charge(robot), 10)
    move_2 = InstantaneousAction('move_2', robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter('robot')
    l_from = move_2.parameter('l_from')
    l_to = move_2.parameter('l_to')
    move_2.add_precondition(GE(battery_charge(robot), 15))
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable('mid_loc', Location)
    #(E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(Exists(And(Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                            Or(is_connected(l_from, mid_location), is_connected(mid_location, l_from)),
                            Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to))), mid_location))
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    move_2.add_decrease_effect(battery_charge(robot), 15)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    l3 = Object('l3', Location)
    l4 = Object('l4', Location)
    l5 = Object('l5', Location)
    r1 = Object('r1', Robot)
    problem = Problem('robot_locations_connected')
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(battery_charge)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.set_initial_value(battery_charge(r1), 100)
    problem.add_goal(is_at(l5, r1))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move_2, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l3))),
                                   up.plan.ActionInstance(move_2, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l5)))])
    robot_locations_connected = Example(problem=problem, plan=plan)
    problems['robot_locations_connected'] = robot_locations_connected

    #robot locations visited
    Location = UserType('Location')
    Robot = UserType('Robot')
    is_at = Fluent('is_at', BoolType(), position=Location, robot=Robot)
    battery_charge = Fluent('battery_charge', RealType(0, 100), robot=Robot)
    is_connected = Fluent('is_connected', BoolType(), location_1=Location, location_2=Location)
    visited = Fluent('visited', BoolType(), target=Location)
    move = InstantaneousAction('move', robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter('robot')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge(robot), 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move.add_effect(visited(l_to), True)
    move.add_decrease_effect(battery_charge(robot), 10)
    move_2 = InstantaneousAction('move_2', robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter('robot')
    l_from = move_2.parameter('l_from')
    l_to = move_2.parameter('l_to')
    move_2.add_precondition(GE(battery_charge(robot), 15))
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable('mid_loc', Location)
    #(E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(Exists(And(Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                            Or(is_connected(l_from, mid_location), is_connected(mid_location, l_from)),
                            Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to))), mid_location))
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    move_2.add_effect(visited(l_to), True)
    move_2.add_decrease_effect(battery_charge(robot), 15)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    l3 = Object('l3', Location)
    l4 = Object('l4', Location)
    l5 = Object('l5', Location)
    r1 = Object('r1', Robot)
    problem = Problem('robot_locations_visited')
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(battery_charge)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_fluent(visited, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(visited(l1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.set_initial_value(battery_charge(r1), 50)
    problem.add_goal(is_at(l5, r1))
    visited_location = Variable('visited_loc', Location)
    problem.add_goal(Forall(visited(visited_location) , visited_location))
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(move, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l2))),
                                   up.plan.ActionInstance(move, (ObjectExp(r1), ObjectExp(l2), ObjectExp(l3))),
                                   up.plan.ActionInstance(move, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l4))),
                                   up.plan.ActionInstance(move, (ObjectExp(r1), ObjectExp(l4), ObjectExp(l5)))])
    robot_locations_visited = Example(problem=problem, plan=plan)
    problems['robot_locations_visited'] = robot_locations_visited

    # charger_discharger
    charger = Fluent('charger')
    b_1 = Fluent('b_1')
    b_2 = Fluent('b_2')
    b_3 = Fluent('b_3')
    charge = InstantaneousAction('charge')
    discharge = InstantaneousAction('discharge')
    charge.add_precondition(Not(charger))
    charge.add_effect(charger, True)
    # !(charger => (b_1 && b_2 && b_3)) in dnf:
    # (charger and !b_1 ) or (charger and !b_2) or (charger and !b_3)
    # which represents the charger is full and at least one battery is not
    discharge.add_precondition(Not(Implies(charger, And(b_1, b_2, b_3))))
    discharge.add_effect(charger, False)
    discharge.add_effect(b_1, True, Not(b_1))
    discharge.add_effect(b_2, True, And(b_1, Not(b_2)))
    discharge.add_effect(b_3, True, And(b_1, b_2, Not(b_3)))
    problem = Problem('charger_discharger')
    problem.add_fluent(charger)
    problem.add_fluent(b_1)
    problem.add_fluent(b_2)
    problem.add_fluent(b_3)
    problem.add_action(charge)
    problem.add_action(discharge)
    problem.set_initial_value(charger, False)
    problem.set_initial_value(b_1, False)
    problem.set_initial_value(b_2, False)
    problem.set_initial_value(b_3, False)
    problem.add_goal(b_1)
    problem.add_goal(b_2)
    problem.add_goal(b_3)
    plan = up.plan.SequentialPlan([up.plan.ActionInstance(charge), up.plan.ActionInstance(discharge),
                                   up.plan.ActionInstance(charge), up.plan.ActionInstance(discharge),
                                   up.plan.ActionInstance(charge), up.plan.ActionInstance(discharge)])
    charge_discharge = Example(problem=problem, plan=plan)
    problems['charge_discharge'] = charge_discharge

    # matchcellar
    Match = UserType('Match')
    Fuse = UserType('Fuse')
    handfree = Fluent('handfree')
    light = Fluent('light')
    match_used = Fluent('match_used', BoolType(), match=Match)
    fuse_mended = Fluent('fuse_mended', BoolType(), fuse=Fuse)
    light_match = DurativeAction('light_match', m=Match)
    m = light_match.parameter('m')
    light_match.set_fixed_duration(6)
    light_match.add_condition(StartTiming(), Not(match_used(m)))
    light_match.add_effect(StartTiming(), match_used(m), True)
    light_match.add_effect(StartTiming(), light, True)
    light_match.add_effect(EndTiming(), light, False)
    mend_fuse = DurativeAction('mend_fuse', f=Fuse)
    f = mend_fuse.parameter('f')
    mend_fuse.set_fixed_duration(5)
    mend_fuse.add_condition(StartTiming(), handfree)
    mend_fuse.add_condition(ClosedTimeInterval(StartTiming(), EndTiming()), light)
    mend_fuse.add_effect(StartTiming(), handfree, False)
    mend_fuse.add_effect(EndTiming(), fuse_mended(f), True)
    mend_fuse.add_effect(EndTiming(), handfree, True)
    f1 = Object('f1', Fuse)
    f2 = Object('f2', Fuse)
    f3 = Object('f3', Fuse)
    m1 = Object('m1', Match)
    m2 = Object('m2', Match)
    m3 = Object('m3', Match)
    problem = Problem('MatchCellar')
    problem.add_fluent(handfree)
    problem.add_fluent(light)
    problem.add_fluent(match_used, default_initial_value=False)
    problem.add_fluent(fuse_mended, default_initial_value=False)
    problem.add_action(light_match)
    problem.add_action(mend_fuse)
    problem.add_object(f1)
    problem.add_object(f2)
    problem.add_object(f3)
    problem.add_object(m1)
    problem.add_object(m2)
    problem.add_object(m3)
    problem.set_initial_value(light, False)
    problem.set_initial_value(handfree, True)
    problem.add_goal(fuse_mended(f1))
    problem.add_goal(fuse_mended(f2))
    problem.add_goal(fuse_mended(f3))
    plan = up.plan.TimeTriggeredPlan([(Fraction(0, 1), up.plan.ActionInstance(light_match, (ObjectExp(m1), )), Fraction(6, 1)),
                                      (Fraction(1, 100), up.plan.ActionInstance(mend_fuse, (ObjectExp(f1), )), Fraction(5, 1)),
                                      (Fraction(601, 100), up.plan.ActionInstance(light_match, (ObjectExp(m2), )), Fraction(6, 1)),
                                      (Fraction(602, 100), up.plan.ActionInstance(mend_fuse, (ObjectExp(f2), )), Fraction(5, 1)),
                                      (Fraction(1202, 100), up.plan.ActionInstance(light_match, (ObjectExp(m3), )), Fraction(6, 1)),
                                      (Fraction(1203, 100), up.plan.ActionInstance(mend_fuse, (ObjectExp(f3), )), Fraction(5, 1))])
    matchcellar = Example(problem=problem, plan=plan)
    problems['matchcellar'] = matchcellar

    # timed connected locations
    Location = UserType('Location')
    is_connected = Fluent('is_connected', BoolType(), location_1=Location, location_2=Location)
    is_at = Fluent('is_at', BoolType(), position=Location)
    move = DurativeAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.set_fixed_duration(6)
    move.add_condition(StartTiming(), is_at(l_from))
    move.add_condition(StartTiming(), Not(is_at(l_to)))
    mid_location = Variable('mid_loc', Location)
    #(E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move.add_condition(ClosedTimeInterval(StartTiming(), EndTiming()), Exists(And(Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                       Or(is_connected(l_from, mid_location), is_connected(mid_location, l_from)),
                       Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to))), mid_location))
    move.add_condition(StartTiming(), Exists(And(Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                       Or(is_connected(l_from, mid_location), is_connected(mid_location, l_from)),
                       Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to))), mid_location))
    move.add_effect(StartTiming(1), is_at(l_from), False)
    move.add_effect(EndTiming(5), is_at(l_to), True)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    l3 = Object('l3', Location)
    l4 = Object('l4', Location)
    l5 = Object('l5', Location)
    problem = Problem('timed_connected_locations')
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.add_goal(is_at(l5))
    plan = up.plan.TimeTriggeredPlan([(Fraction(0, 1), up.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l3))), Fraction(6, 1)),
                                      (Fraction(6, 1), up.plan.ActionInstance(move, (ObjectExp(l3), ObjectExp(l5))), Fraction(6, 1))])
    timed_connected_locations = Example(problem=problem, plan=plan)
    problems['timed_connected_locations'] = timed_connected_locations

    # hierarchical blocks world
    Entity = UserType('Entity', None) # None can be avoided
    Location = UserType('Location', Entity)
    Unmovable = UserType('Unmovable', Location)
    TableSpace = UserType('TableSpace', Unmovable)
    Movable = UserType('Movable', Location)
    Block = UserType('Block', Movable)
    clear = Fluent('clear', BoolType(), space=Location)
    on = Fluent('on', BoolType(), object=Movable, space=Location)

    move = InstantaneousAction('move', item=Movable, l_from=Location, l_to=Location)
    item = move.parameter('item')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem('hierarchical_blocks_world')
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object('ts_1', TableSpace)
    ts_2 = Object('ts_2', TableSpace)
    ts_3 = Object('ts_3', TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object('block_1', Block)
    block_2 = Object('block_2', Block)
    block_3 = Object('block_3', Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = up.plan.SequentialPlan([
            up.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))),
            up.plan.ActionInstance(move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))),
            up.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))),
            up.plan.ActionInstance(move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2)))])
    hierarchical_blocks_world = Example(problem=problem, plan=plan)
    problems['hierarchical_blocks_world'] = hierarchical_blocks_world

    return problems



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
'''This module defines the different plan classes.'''


import unified_planning as up
import unified_planning.model
from typing import Optional, Tuple, List
from fractions import Fraction


class Plan:
    '''Represents a generic plan.'''
    pass


class ActionInstance:
    '''Represents an action instance with the actual parameters.'''
    def __init__(self, action: 'unified_planning.model.Action', params: Tuple['unified_planning.model.FNode', ...] = tuple()):
        assert len(action.parameters) == len(params)
        self._action = action
        self._params = tuple(params)

    def __repr__(self) -> str:
        s = []
        if len(self._params) > 0:
            s.append('(')
            first = True
            for p in self._params:
                if not first:
                    s.append(', ')
                s.append(str(p))
                first = False
            s.append(')')
        return self._action.name + ''.join(s)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, ActionInstance):
            return self.action == oth.action and self.actual_parameters == oth.actual_parameters
        else:
            return False

    @property
    def action(self) -> 'unified_planning.model.Action':
        '''Returns the action.'''
        return self._action

    @property
    def actual_parameters(self) -> Tuple['unified_planning.model.FNode', ...]:
        '''Returns the actual parameters.'''
        return self._params


class SequentialPlan(Plan):
    '''Represents a sequential plan.'''
    def __init__(self, actions: List[ActionInstance]):
        self._actions = actions

    def __repr__(self) -> str:
        return str(self._actions)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, SequentialPlan):
            return self.actions == oth.actions
        else:
            return False

    @property
    def actions(self) -> List[ActionInstance]:
        '''Returns the sequence of action instances.'''
        return self._actions


class TimeTriggeredPlan(Plan):
    '''Represents a time triggered plan.'''
    def __init__(self, actions: List[Tuple[Fraction, ActionInstance, Optional[Fraction]]]):
        '''The first Fraction represents the absolute time in which the action
        Action starts, while the last Fraction represents the duration
        of the action to fullfill the problem goals.
        The Action can be an InstantaneousAction, this is represented with a duration set
        to None.
        '''
        self._actions = actions

    def __repr__(self) -> str:
        return str(self._actions)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, TimeTriggeredPlan):
            return self.actions == oth.actions
        else:
            return False

    @property
    def actions(self) -> List[Tuple[Fraction, ActionInstance, Optional[Fraction]]]:
        '''Returns the sequence of action instances.'''
        return self._actions



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.model.fnode import FNode
from enum import Enum, auto
from fractions import Fraction
from typing import Union


class TimepointKind(Enum):
    GLOBAL_START = auto()
    GLOBAL_END = auto()
    START = auto()
    END = auto()


class Timepoint:
    def __init__(self, kind: TimepointKind):
        self._kind = kind

    def __repr__(self):
        if self._kind == TimepointKind.GLOBAL_START or self._kind == TimepointKind.START:
            return 'start'
        else:
            return 'end'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Timepoint):
            return self._kind == oth._kind
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._kind)

    @property
    def kind(self) -> TimepointKind:
        return self._kind


class Timing:
    def __init__(self, delay: Union[int, Fraction], timepoint: Timepoint):
        self._timepoint = timepoint
        self._delay = delay

    def __repr__(self):
        if self._delay == 0:
            return f'{self._timepoint}'
        else:
            return f'{self._timepoint} + {self._delay}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Timing):
            return self._delay == oth._delay and self._timepoint == oth._timepoint
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._delay) ^ hash(self._timepoint)

    @property
    def delay(self) -> Union[int, Fraction]:
        return self._delay

    def is_global(self) -> bool:
        return self._timepoint.kind == TimepointKind.GLOBAL_START or \
            self._timepoint.kind == TimepointKind.GLOBAL_END

    def is_from_start(self) -> bool:
        return self._timepoint.kind == TimepointKind.START or \
            self._timepoint.kind == TimepointKind.GLOBAL_START

    def is_from_end(self) -> bool:
        return not self.is_from_start()


def StartTiming(delay: Union[int, Fraction] = 0) -> Timing:
    '''Represents the start timing of an action.
    Created with a delay > 0 represents "delay" time
    after the start of an action.

    For example, action starts at time 5:
    StartTiming() = 5
    StartTiming(3) = 5+3 = 8'''

    return Timing(delay, Timepoint(TimepointKind.START))


def EndTiming(delay: Union[int, Fraction] = 0) -> Timing:
    '''Represents the end timing of an action.
    Created with a delay > 0 represents "delay" time
    before the end of an action.

    For example, action ends at time 10:
    EndTiming() = 10
    EndTiming(1.5) = 10-Fraction(3, 2) = Fraction(17, 2) = 8.5'''

    return Timing(delay, Timepoint(TimepointKind.END))


def GlobalStartTiming(delay: Union[int, Fraction] = 0):
    '''Represents the absolute timing.
    Created with a delay > 0 represents "delay" time
    after the start of the execution.'''

    return Timing(delay, Timepoint(TimepointKind.GLOBAL_START))


def GlobalEndTiming(delay: Union[int, Fraction] = 0):
    '''Represents the end timing of an execution.
    Created with a delay > 0 represents "delay" time
    before the end of the execution.'''

    return Timing(delay, Timepoint(TimepointKind.GLOBAL_END))


class Interval:
    def __init__(self, lower: FNode, upper: FNode, is_left_open: bool = False, is_right_open: bool = False):
        self._lower = lower
        self._upper = upper
        self._is_left_open = is_left_open
        self._is_right_open = is_right_open

    def __repr__(self) -> str:
        if self.is_left_open():
            left_bound = '('
        else:
            left_bound = '['
        if self.is_right_open():
            right_bound = ')'
        else:
            right_bound = ']'
        return f'{left_bound}{str(self.lower)}, {str(self.upper)}{right_bound}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Interval):
            return self._lower == oth._lower and self._upper == oth._upper and self._is_left_open == oth._is_left_open and self._is_right_open == oth._is_right_open
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._lower) + hash(self._upper)
        if self._is_left_open:
            res ^= hash('is_left_open')
        if self._is_right_open:
            res ^= hash('is_right_open')
        return res

    @property
    def lower(self) -> FNode:
        return self._lower

    @property
    def upper(self) -> FNode:
        return self._upper

    def is_left_open(self) -> bool:
        return self._is_left_open

    def is_right_open(self) -> bool:
        return self._is_right_open


class Duration:
    pass


class DurationInterval(Duration, Interval):
    def __init__(self, lower: FNode, upper: FNode, is_left_open: bool = False, is_right_open: bool = False):
        Duration.__init__(self)
        Interval.__init__(self, lower, upper, is_left_open, is_right_open)


def ClosedDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    '''Represents the (closed) interval duration constraint:
            [lower, upper]
    '''
    return DurationInterval(lower, upper)


def FixedDuration(size: FNode) -> DurationInterval:
    '''Represents a fixed duration constraint'''
    return DurationInterval(size, size)


def OpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    '''Represents the (open) interval duration constraint:
            (lower, upper)
    '''
    return DurationInterval(lower, upper, True, True)


def LeftOpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    '''Represents the (left open, right closed) interval duration constraint:
            (lower, upper]
    '''
    return DurationInterval(lower, upper, True, False)


def RightOpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    '''Represents the (left closed, right open) interval duration constraint:
            [lower, upper)
    '''
    return DurationInterval(lower, upper, False, True)


class TimeInterval:
    def __init__(self, lower: Timing, upper: Timing, is_left_open: bool = False, is_right_open: bool = False):
        self._lower = lower
        self._upper = upper
        self._is_left_open = is_left_open
        self._is_right_open = is_right_open

    def __repr__(self) -> str:
        if self.is_left_open():
            left_bound = '('
        else:
            left_bound = '['
        if self.is_right_open():
            right_bound = ')'
        else:
            right_bound = ']'
        if self.lower == self.upper:
            return f'{left_bound}{str(self.lower)}{right_bound}'
        else:
            return f'{left_bound}{str(self.lower)}, {str(self.upper)}{right_bound}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, TimeInterval):
            return self._lower == oth._lower and self._upper == oth._upper and self._is_left_open == oth._is_left_open and self._is_right_open == oth._is_right_open
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._lower) + hash(self._upper)
        if self._is_left_open:
            res ^= hash('is_left_open')
        if self._is_right_open:
            res ^= hash('is_right_open')
        return res

    @property
    def lower(self) -> Timing:
        return self._lower

    @property
    def upper(self) -> Timing:
        return self._upper

    def is_left_open(self) -> bool:
        return self._is_left_open

    def is_right_open(self) -> bool:
        return self._is_right_open


def TimePointInterval(tp: Timing) -> TimeInterval:
    '''Represents the (point) temporal interval: [tp, tp]'''
    return TimeInterval(tp, tp)


def ClosedTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    '''Represents the (closed) temporal interval: [lower, upper]'''
    return TimeInterval(lower, upper)


def OpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    '''Represents the (open) temporal interval: (lower, upper)'''
    return TimeInterval(lower, upper, True, True)


def LeftOpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    '''Represents the (left open, right closed) temporal interval: (lower, upper]'''
    return TimeInterval(lower, upper, True, False)


def RightOpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    '''Represents the (left closed, right open) temporal interval: [lower, upper)'''
    return TimeInterval(lower, upper, False, True)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines an Object of a planning problem.
An Object is represented by a name and by its type.
"""

import unified_planning.model.types


class Object:
    """Represents an object."""
    def __init__(self, name: str, typename: 'unified_planning.model.types.Type'):
        self._name = name
        self._typename = typename

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Object):
            return self._name == oth._name and self._typename == oth._typename
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + hash(self._typename)

    @property
    def name(self) -> str:
        """Returns the object name."""
        return self._name

    @property
    def type(self) -> 'unified_planning.model.types.Type':
        """Returns the object type."""
        return self._typename



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unified_planning.walkers.generic import Walker
from unified_planning.model.fnode import FNode


class DagWalker(Walker):
    """DagWalker treats the expression as a DAG and performs memoization of the
    intermediate results.
    This should be used when the result of applying the function to a
    expression is always the same, independently of where the expression has
    been found; examples include substitution and solving.
    Due to memoization, a few more things need to be taken into
    account when using the DagWalker.
    :func _get_key needs to be defined if additional arguments via
    keywords need to be shared. This function should return the key to
    be used in memoization. See substituter for an example.
    """

    def __init__(self, invalidate_memoization=False):
        """The flag ``invalidate_memoization`` can be used to clear the cache
        after the walk has been completed: the cache is one-time use.
        """
        Walker.__init__(self)

        self.memoization = {}
        self.invalidate_memoization = invalidate_memoization
        self.stack = []
        return

    def _get_children(self, expression: FNode):
        return expression.args

    def _push_with_children_to_stack(self, expression: FNode, **kwargs):
        """Add children to the stack."""
        self.stack.append((True, expression))
        for s in self._get_children(expression):
            # Add only if not memoized already
            key = self._get_key(s, **kwargs)
            if key not in self.memoization:
                self.stack.append((False, s))

    def _compute_node_result(self, expression: FNode, **kwargs):
        """Apply function to the node and memoize the result.
        Note: This function assumes that the results for the children
              are already available.
        """
        key = self._get_key(expression, **kwargs)
        if key not in self.memoization:
            try:
                f = self.functions[expression.node_type]
            except KeyError:
                f = self.walk_error

            args = [self.memoization[self._get_key(s, **kwargs)] \
                    for s in self._get_children(expression)]
            self.memoization[key] = f(expression, args=args, **kwargs)
        else:
            pass

    def _process_stack(self, **kwargs):
        """Empties the stack by processing every node in it.
        Processing is performed in two steps.
        1- A node is expanded and all its children are push to the stack
        2- Once all children have been processed, the result for the node
           is computed and memoized.
        """

        while self.stack:
            (was_expanded, expression) = self.stack.pop()
            if was_expanded:
                self._compute_node_result(expression, **kwargs)
            else:
                self._push_with_children_to_stack(expression, **kwargs)

    def iter_walk(self, expression: FNode, **kwargs):
        """Performs an iterative walk of the DAG"""
        self.stack.append((False, expression))
        self._process_stack(**kwargs)
        res_key = self._get_key(expression, **kwargs)
        return self.memoization[res_key]

    def walk(self, expression: FNode, **kwargs):
        if expression in self.memoization:
            return self.memoization[expression]

        res = self.iter_walk(expression, **kwargs)

        if self.invalidate_memoization:
            self.memoization.clear()
        return res

    def _get_key(self, expression: FNode, **kwargs):
        if not kwargs:
            return expression
        raise NotImplementedError("DagWalker should redefine '_get_key'" +
                                  " when using keywords arguments")

    def walk_true(self, expression: FNode, args, **kwargs):
        #pylint: disable=unused-argument
        """ Returns True, independently from the children's value."""
        return True

    def walk_false(self, expression: FNode, args, **kwargs):
        #pylint: disable=unused-argument
        """ Returns False, independently from the children's value."""
        return False

    def walk_none(self, expression: FNode, args, **kwargs):
        #pylint: disable=unused-argument
        """ Returns None, independently from the children's value."""
        return None

    def walk_identity(self, expression: FNode, **kwargs):
        #pylint: disable=unused-argument
        """ Returns expression, independently from the childrens's value."""
        return expression

    def walk_any(self, expression: FNode, args, **kwargs):
        #pylint: disable=unused-argument
        """ Returns True if any of the children returned True. """
        return any(args)

    def walk_all(self, expression: FNode, args, **kwargs):
        #pylint: disable=unused-argument
        """ Returns True if all the children returned True. """
        return all(args)

# EOC DagWalker



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
import warnings
import unified_planning as up
import unified_planning.solvers as solvers
from unified_planning.plan import Plan, ActionInstance
from unified_planning.model import ProblemKind
from unified_planning.exceptions import UPException
from unified_planning.solvers.results import LogLevel, PlanGenerationResultStatus
from typing import IO, Callable, Dict, List, Optional, Tuple, cast
from multiprocessing import Process, Queue


class Parallel(solvers.solver.Solver):
    """Create a parallel instance of multiple Solvers."""

    def __init__(self, solvers: List[Tuple[type, Dict[str, str]]]):
        self.solvers = solvers

    @property
    def name(self) -> str:
        return 'Parallel'

    @staticmethod
    def is_oneshot_planner() -> bool:
        raise UPException('The Parallel solver type depends on its actual solvers')

    @staticmethod
    def is_plan_validator() -> bool:
        raise UPException('The Parallel solver type depends on its actual solvers')

    @staticmethod
    def supports(problem_kind: 'ProblemKind') -> bool:
        raise UPException('The Parallel supported features depends on its actual solvers')

    def _run_parallel(self, fname, *args):
        signaling_queue = Queue()
        processes = []
        for idx, (solver_class, opts) in enumerate(self.solvers):
            options = opts
            _p = Process(name=str(idx),
                         target=_run,
                         args=(idx, solver_class, options,
                               signaling_queue, fname, *args))
            processes.append(_p)
            _p.start()
        processes_alive = len(processes)
        results: List[up.solvers.PlanGenerationResult] = []
        planners_final_status: List[str] = []
        definitive_result_found: bool = False
        optimality_required: bool = len(args[0].quality_metrics) > 0 # Require optimality if the problem has at least one quality metric.
        while True:
            if processes_alive == 0: # Every planner gave a result
                break
            (idx, res) = signaling_queue.get(block=True)
            processes_alive -= 1
            if isinstance(res, BaseException):
                raise res
            else:
                assert isinstance(res, up.solvers.PlanGenerationResult)
                # If the planner is sure about the result (optimality of the result or impossibility of the problem or the problem does not need optimality) exit the loop
                if res.status == PlanGenerationResultStatus.SOLVED_OPTIMALLY or res.status == PlanGenerationResultStatus.UNSOLVABLE_PROVEN or (res.status == PlanGenerationResultStatus.SOLVED_SATISFICING and not optimality_required):
                    definitive_result_found = True
                    break
                else:
                    planners_final_status.append(f'{res.planner_name}: {res.status_as_str()}')
                    if res.plan is not None:
                        results.append(res)
        for p in processes:
            p.terminate()
        if definitive_result_found: # A planner found a definitive result
            return res
        else:
            result_order: List[int] = [  PlanGenerationResultStatus.SOLVED_SATISFICING,  # List containing the results in the order we prefer them
                                         PlanGenerationResultStatus.TIMEOUT,
                                         PlanGenerationResultStatus.MEMOUT,
                                         PlanGenerationResultStatus.INTERNAL_ERROR]
            logs = up.solvers.LogMessage(LogLevel.INFO, ', '.join(planners_final_status))
            for ro in result_order:
                for r in results:
                    if r.status == ro:
                        return r # NOTE Here we may want to append the logs variable to log_messages
            # if no results are given by the planner, we create a default one
            return up.solvers.PlanGenerationResult(PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY,
                                                    None, self.name, log_messages=[logs])


    def solve(self, problem: 'up.model.Problem',
                    callback: Optional[Callable[['up.solvers.results.PlanGenerationResult'], None]] = None,
                    timeout: Optional[float] = None,
                    output_stream: Optional[IO[str]] = None) -> 'up.solvers.results.PlanGenerationResult':
        if callback is not None:
            warnings.warn('Parallel solvers do not support the callback system.', UserWarning)
        if output_stream is not None:
            warnings.warn('Parallel solvers do not support the output stream system.', UserWarning)
        final_report = self._run_parallel('solve', problem, None, timeout, None)
        new_plan = self.convert_plan(final_report.plan, problem)
        return up.solvers.results.PlanGenerationResult(final_report.status, new_plan, final_report.planner_name, final_report.metrics, final_report.log_messages)

    def convert_plan(self, plan: 'up.plan.Plan', problem: 'up.model.Problem')-> 'up.plan.Plan':
        objects = {}
        for ut in problem.user_types:
            for obj in problem.objects(ut):
                objects[obj.name] = obj
        em = problem.env.expression_manager
        actions: List[ActionInstance] = []
        if isinstance(plan, up.plan.SequentialPlan):
            actions = plan.actions
        elif isinstance(plan, up.plan.TimeTriggeredPlan):
            actions = [a for _, a, _ in plan.actions]
        else:
            raise NotImplementedError
        new_actions: List[ActionInstance] = []
        for a in actions:
            new_a = problem.action(a.action.name)
            params = []
            for p in a.actual_parameters:
                if p.is_object_exp():
                    obj = objects[p.object().name]
                    params.append(em.ObjectExp(obj))
                elif p.is_bool_constant():
                    params.append(em.Bool(p.is_true()))
                elif p.is_int_constant():
                    params.append(em.Int(cast(int, p.constant_value())))
                elif p.is_real_constant():
                    params.append(em.Real(cast(Fraction, p.constant_value())))
                else:
                    raise
            new_actions.append(ActionInstance(new_a, tuple(params)))
        if isinstance(plan, up.plan.SequentialPlan):
            return up.plan.SequentialPlan(new_actions)
        elif isinstance(plan, up.plan.TimeTriggeredPlan):
            return up.plan.TimeTriggeredPlan([(t, a, d) for (t, _, d), a in zip(plan.actions, new_actions)])
        else:
            raise NotImplementedError

    def validate(self, problem: 'up.model.Problem', plan: Plan) -> bool:
        return self._run_parallel('validate', problem, plan)

    def destroy(self):
        pass


def _run(idx: int, SolverClass: type, options: Dict[str, str], signaling_queue: Queue, fname: str, *args):
    with SolverClass(**options) as s:
        try:
            local_res = getattr(s, fname)(*args)
        except Exception as ex:
            signaling_queue.put((idx, ex))
            return
        signaling_queue.put((idx, local_res))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import partialmethod, total_ordering
from typing import Set


# TODO: This features map needs to be extended with all the problem characterizations.
FEATURES = {
    'TIME' : ['CONTINUOUS_TIME', 'DISCRETE_TIME', 'INTERMEDIATE_CONDITIONS_AND_EFFECTS', 'TIMED_EFFECT', 'TIMED_GOALS', 'DURATION_INEQUALITIES'],
    'NUMBERS' : ['CONTINUOUS_NUMBERS', 'DISCRETE_NUMBERS'],
    'CONDITIONS_KIND' : ['NEGATIVE_CONDITIONS', 'DISJUNCTIVE_CONDITIONS', 'EQUALITY', 'EXISTENTIAL_CONDITIONS', 'UNIVERSAL_CONDITIONS'],
    'EFFECTS_KIND' : ['CONDITIONAL_EFFECTS', 'INCREASE_EFFECTS', 'DECREASE_EFFECTS'],
    'TYPING' : ['FLAT_TYPING', 'HIERARCHICAL_TYPING'],
    'FLUENTS_TYPE' : ['NUMERIC_FLUENTS', 'OBJECT_FLUENTS'],
    'QUALITY_METRICS' : ['ACTIONS_COST', 'FINAL_VALUE', 'MAKESPAN', 'PLAN_LENGTH']
}


class ProblemKindMeta(type):
    '''Meta class used to interpret the nodehandler decorator.'''
    def __new__(cls, name, bases, dct):
        def _set(self, feature, possible_features):
            assert feature in possible_features
            self._features.add(feature)

        def _has(self, feature):
            return feature in self._features

        obj = type.__new__(cls, name, bases, dct)
        for m, l in FEATURES.items():
            setattr(obj, "set_" + m.lower(), partialmethod(_set, possible_features=l))
            for f in l:
                setattr(obj, "has_" + f.lower(), partialmethod(_has, feature=f))
        return obj

@total_ordering
class ProblemKind(metaclass=ProblemKindMeta):
    def __init__(self, features: Set[str] = set()):
        self._features: Set[str] = set(features)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, ProblemKind):
            return self._features == oth._features
        else:
            return False

    def __hash__(self) -> int:
        res = 0
        for f in self._features:
            res += hash(f)
        return res

    def __le__(self, oth: object):
        if not isinstance(oth, ProblemKind):
            raise
        return self._features.issubset(oth._features)

    @property
    def features(self) -> Set[str]:
        return self._features

    def union(self, oth: 'ProblemKind') -> 'ProblemKind':
        return ProblemKind(self.features.union(oth.features))


basic_classical_kind = ProblemKind()
basic_classical_kind.set_typing('FLAT_TYPING') # type: ignore

hierarchical_kind = ProblemKind()
hierarchical_kind.set_typing('HIERARCHICAL_TYPING') # type: ignore

classical_kind = ProblemKind()
classical_kind.set_typing('FLAT_TYPING') # type: ignore
classical_kind.set_conditions_kind('NEGATIVE_CONDITIONS') # type: ignore
classical_kind.set_conditions_kind('DISJUNCTIVE_CONDITIONS') # type: ignore
classical_kind.set_conditions_kind('EQUALITY') # type: ignore

full_classical_kind = ProblemKind()
full_classical_kind.set_typing('FLAT_TYPING') # type: ignore
full_classical_kind.set_conditions_kind('NEGATIVE_CONDITIONS') # type: ignore
full_classical_kind.set_conditions_kind('DISJUNCTIVE_CONDITIONS') # type: ignore
full_classical_kind.set_conditions_kind('EQUALITY') # type: ignore
full_classical_kind.set_conditions_kind('EXISTENTIAL_CONDITIONS') # type: ignore
full_classical_kind.set_conditions_kind('UNIVERSAL_CONDITIONS') # type: ignore
full_classical_kind.set_effects_kind('CONDITIONAL_EFFECTS') # type: ignore

object_fluent_kind = ProblemKind()
object_fluent_kind.set_fluents_type('OBJECT_FLUENTS') # type: ignore

basic_numeric_kind = ProblemKind()
basic_numeric_kind.set_typing('FLAT_TYPING') # type: ignore
basic_numeric_kind.set_numbers('DISCRETE_NUMBERS') # type: ignore
basic_numeric_kind.set_numbers('CONTINUOUS_NUMBERS') # type: ignore
basic_numeric_kind.set_fluents_type('NUMERIC_FLUENTS') # type: ignore

full_numeric_kind = ProblemKind()
full_numeric_kind.set_typing('FLAT_TYPING') # type: ignore
full_numeric_kind.set_numbers('DISCRETE_NUMBERS') # type: ignore
full_numeric_kind.set_numbers('CONTINUOUS_NUMBERS') # type: ignore
full_numeric_kind.set_fluents_type('NUMERIC_FLUENTS') # type: ignore
full_numeric_kind.set_effects_kind('INCREASE_EFFECTS') # type: ignore
full_numeric_kind.set_effects_kind('DECREASE_EFFECTS') # type: ignore

basic_temporal_kind = ProblemKind()
basic_temporal_kind.set_typing('FLAT_TYPING') # type: ignore
basic_temporal_kind.set_time('CONTINUOUS_TIME') # type: ignore

full_temporal_kind = ProblemKind()
full_temporal_kind.set_typing('FLAT_TYPING') # type: ignore
full_temporal_kind.set_time('CONTINUOUS_TIME') # type: ignore
full_temporal_kind.set_time('INTERMEDIATE_CONDITIONS_AND_EFFECTS') # type: ignore
full_temporal_kind.set_time('TIMED_EFFECT') # type: ignore
full_temporal_kind.set_time('TIMED_GOALS') # type: ignore
full_temporal_kind.set_time('DURATION_INEQUALITIES') # type: ignore

quality_metrics_kind = ProblemKind()
quality_metrics_kind.set_quality_metrics('ACTIONS_COST') # type: ignore
quality_metrics_kind.set_quality_metrics('FINAL_VALUE') # type: ignore



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import unified_planning as up
from functools import wraps
from unified_planning.environment import get_env
from unified_planning.model import ProblemKind
from unified_planning.test.pddl import enhsp


skipIf = unittest.skipIf
SkipTest = unittest.SkipTest


class skipIfSolverNotAvailable(object):
    """Skip a test if the given solver is not available."""

    def __init__(self, solver):
        self.solver = solver

    def __call__(self, test_fun):
        msg = "%s not available" % self.solver
        cond = self.solver not in get_env().factory.solvers
        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)
        return wrapper


class skipIfNoOneshotPlannerForProblemKind(object):
    """Skip a test if there are no oneshot planner for the given problem kind."""

    def __init__(self, kind: ProblemKind):
        self.kind = kind

    def __call__(self, test_fun):
        msg = "no oneshot planner available for the given problem kind"
        cond = get_env().factory._get_solver_class('oneshot_planner', problem_kind=self.kind) is None
        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)
        return wrapper


class skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee(object):
    """Skip a test if there are no oneshot planner satisfies optimality guarantee."""

    def __init__(self, optimality_guarantee: up.solvers.results.PlanGenerationResultStatus):
        self.optimality_guarantee = optimality_guarantee

    def __call__(self, test_fun):
        msg = "no oneshot planner available for the given optimality guarantee"
        cond = get_env().factory._get_solver_class('oneshot_planner',
                                                   optimality_guarantee=self.optimality_guarantee) is None
        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)
        return wrapper


class skipIfNoPlanValidatorForProblemKind(object):
    """Skip a test if there are no plan validator for the given problem kind."""

    def __init__(self, kind: ProblemKind):
        self.kind = kind

    def __call__(self, test_fun):
        msg = "no plan validator available for the given problem kind"
        cond = get_env().factory._get_solver_class('plan_validator', problem_kind=self.kind) is None
        @unittest.skipIf(cond, msg)
        @wraps(test_fun)
        def wrapper(*args, **kwargs):
            return test_fun(*args, **kwargs)
        return wrapper

TestCase = unittest.TestCase
main = unittest.main



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unified_planning
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import classical_kind, basic_numeric_kind
from unified_planning.test import TestCase, main, skipIfNoPlanValidatorForProblemKind
from unified_planning.test.examples import get_example_problems


class TestPlanValidator(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoPlanValidatorForProblemKind(classical_kind)
    def test_basic(self):
        problem, plan = self.problems['basic']

        with PlanValidator(problem_kind=problem.kind) as validator:
            self.assertNotEqual(validator, None)

            res = validator.validate(problem, plan)
            self.assertTrue(res)

            plan = unified_planning.plan.SequentialPlan([])
            res = validator.validate(problem, plan)
            self.assertFalse(res)

    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_numeric_kind))
    def test_robot(self):
        problem, plan = self.problems['robot']

        with PlanValidator(problem_kind=problem.kind) as validator:
            self.assertNotEqual(validator, None)

            res = validator.validate(problem, plan)
            self.assertTrue(res)

            plan = unified_planning.plan.SequentialPlan([])
            res = validator.validate(problem, plan)
            self.assertFalse(res)

    @skipIfNoPlanValidatorForProblemKind(classical_kind)
    def test_robot_loader(self):
        problem, plan = self.problems['robot_loader']

        with PlanValidator(problem_kind=problem.kind) as validator:
            self.assertNotEqual(validator, None)

            res = validator.validate(problem, plan)
            self.assertTrue(res)

            plan = unified_planning.plan.SequentialPlan([])
            res = validator.validate(problem, plan)
            self.assertFalse(res)

    @skipIfNoPlanValidatorForProblemKind(classical_kind)
    def test_robot_loader_adv(self):
        problem, plan = self.problems['robot_loader_adv']

        with PlanValidator(problem_kind=problem.kind) as validator:
            self.assertNotEqual(validator, None)

            res = validator.validate(problem, plan)
            self.assertTrue(res)

            plan = unified_planning.plan.SequentialPlan([])
            res = validator.validate(problem, plan)
            self.assertFalse(res)


if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Variable class.
A Variable has a name and a type.
"""


from typing import List, Set
from unified_planning.model.fnode import FNode
from unified_planning.model.operators import OperatorKind
import unified_planning
import unified_planning.walkers as walkers
import unified_planning.model.operators as op



class Variable:
    """Represents a varible."""
    def __init__(self, name: str, typename: 'unified_planning.model.types.Type'):
        self._name = name
        self._typename = typename

    def __repr__(self) -> str:
        return f'{str(self.type)} {self.name}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Variable):
            return self._name == oth._name and self._typename == oth._typename
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + hash(self._typename)

    @property
    def name(self) -> str:
        """Returns the variable name."""
        return self._name

    @property
    def type(self) -> 'unified_planning.model.types.Type':
        """Returns the variable type."""
        return self._typename

class FreeVarsOracle(walkers.DagWalker):
    # We have only few categories for this walker.
    #
    # - Quantifiers need to exclude bounded variables
    # - Other operators need to return the union of all their sons
    # - Constants have no impact

    def get_free_variables(self, expression: FNode) -> Set[Variable]:
        """Returns the set of Symbols appearing free in the expression."""
        return self.walk(expression)

    @walkers.handles(OperatorKind.VARIABLE_EXP)
    def walk_variable_exp(self, expression: FNode, args: List[Set[Variable]], **kwargs) -> Set[Variable]:
        #pylint: disable=unused-argument
        return {expression.variable()}

    @walkers.handles(OperatorKind.EXISTS, OperatorKind.FORALL)
    def walk_quantifier(self, expression: FNode, args: List[Set[Variable]], **kwargs) -> Set[Variable]:
        #pylint: disable=unused-argument
        return args[0].difference(expression.variables())

    @walkers.handles(op.CONSTANTS)
    def walk_constant(self, expression: FNode, args: List[Set[Variable]], **kwargs) -> Set[Variable]:
        #pylint: disable=unused-argument
        return set()

    @walkers.handles(set(OperatorKind) - {OperatorKind.VARIABLE_EXP, OperatorKind.EXISTS, OperatorKind.FORALL})
    def walk_all(self, expression: FNode, args: List[Set[Variable]], **kwargs) -> Set[Variable]:
        return {v for s in args for v in s}



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from io import StringIO
import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.solvers import PlanGenerationResultStatus
from unified_planning.test import TestCase, main, skipIfSolverNotAvailable
from unified_planning.test.examples import get_example_problems

VERYSMALL_TIMEOUT=0.0001

class TestPDDLPlanner(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_basic(self):
        problem = self.problems['basic'].problem
        a = problem.action('a')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_basic_conditional(self):
        problem = self.problems['basic_conditional'].problem
        a_x = problem.action('a_x')
        a_y = problem.action('a_y')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 2)
            self.assertEqual(plan.actions[0].action, a_y)
            self.assertEqual(plan.actions[1].action, a_x)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)
            self.assertEqual(len(plan.actions[1].actual_parameters), 0)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot(self):
        problem = self.problems['robot'].problem
        move = problem.action('move')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_decrease(self):
        problem = self.problems['robot_decrease'].problem
        move = problem.action('move')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader(self):
        problem = self.problems['robot_loader'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 4)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)
            self.assertEqual(len(plan.actions[1].actual_parameters), 1)
            self.assertEqual(len(plan.actions[2].actual_parameters), 2)
            self.assertEqual(len(plan.actions[3].actual_parameters), 1)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader_adv(self):
        problem = self.problems['robot_loader_adv'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')

        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)
            output = StringIO()
            final_report = planner.solve(problem, output_stream=output)
            plan = final_report.plan
            planner_output = output.getvalue()
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

            self.assertIn('Domain parsed\nProblem parsed\n', planner_output)
            self.assertIn('\nGrounding..\nGrounding Time:', planner_output)
            self.assertIn('Aibr Preprocessing\n|F|:7\n', planner_output)
            self.assertIn('\n|X|:0\n|A|:15\n', planner_output)
            self.assertIn('\n|P|:0\n|E|:0\nH1 Setup Time (msec): ', planner_output)
            self.assertIn('Setting horizon to:NaN\nHelpful Action Pruning Activated\n', planner_output)
            self.assertIn('\nRunning WA-STAR\nh(n = s_0)=3.0\n', planner_output)
            self.assertIn('f(n) = 3.0 (Expanded Nodes: 0, Evaluated States: 0, Time: ', planner_output)
            self.assertIn('f(n) = 4.0 (Expanded Nodes: 2, Evaluated States: 3, Time: ', planner_output)
            self.assertIn('f(n) = 5.0 (Expanded Nodes: 5, Evaluated States: 6, Time: ', planner_output)
            self.assertIn('Problem Solved\n\nFound Plan:\n', planner_output)
            self.assertIn('\n0.0: (move l1 l2 r1)\n1.0: (load l2 r1 c1)\n2.0: (move l2 l3 r1)', planner_output)
            self.assertIn('3.0: (unload l3 r1 c1)\n4.0: (move l3 l1 r1)\n\nPlan-Length:5\n', planner_output)
            self.assertIn('\nMetric (Search):5.0\nPlanning Time (msec):', planner_output)
            self.assertIn('\nHeuristic Time (msec): ', planner_output)
            self.assertIn('\nSearch Time (msec): ', planner_output)
            self.assertIn('\nExpanded Nodes:7\nStates Evaluated:8\n', planner_output)
            self.assertIn('\nFixed constraint violations during search (zero-crossing):0\nNumber of Dead-Ends detected:0\n', planner_output)
            self.assertIn('\nNumber of Duplicates detected:8\n', planner_output)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader_adv_with_timeout(self):
        problem, right_plan = self.problems['robot_loader_adv'].problem, self.problems['robot_loader_adv'].plan
        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout = VERYSMALL_TIMEOUT)
            self.assertIn(final_report.status, [PlanGenerationResultStatus.TIMEOUT, PlanGenerationResultStatus.SOLVED_OPTIMALLY]) # It could happen that the PDDL solver manages to solve the problem
            self.assertTrue(final_report.plan is None or final_report.plan == right_plan)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader_adv_with_long_timeout(self):
        problem = self.problems['robot_loader_adv'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')
        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout=100)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader_adv_with_long_timeout_and_output_stream(self):
        problem = self.problems['robot_loader_adv'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')
        output_stream = StringIO()
        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout=100, output_stream=output_stream)
            plan = final_report.plan
            planner_output = output_stream.getvalue()
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

            self.assertIn('Domain parsed\nProblem parsed\n', planner_output)
            self.assertIn('\nGrounding..\nGrounding Time:', planner_output)
            self.assertIn('Aibr Preprocessing\n|F|:7\n', planner_output)
            self.assertIn('\n|X|:0\n|A|:15\n', planner_output)
            self.assertIn('\n|P|:0\n|E|:0\nH1 Setup Time (msec): ', planner_output)
            self.assertIn('Setting horizon to:NaN\nHelpful Action Pruning Activated\n', planner_output)
            self.assertIn('\nRunning WA-STAR\nh(n = s_0)=3.0\n', planner_output)
            self.assertIn('f(n) = 3.0 (Expanded Nodes: 0, Evaluated States: 0, Time: ', planner_output)
            self.assertIn('f(n) = 4.0 (Expanded Nodes: 2, Evaluated States: 3, Time: ', planner_output)
            self.assertIn('f(n) = 5.0 (Expanded Nodes: 5, Evaluated States: 6, Time: ', planner_output)
            self.assertIn('Problem Solved\n\nFound Plan:\n', planner_output)
            self.assertIn('\n0.0: (move l1 l2 r1)\n1.0: (load l2 r1 c1)\n2.0: (move l2 l3 r1)', planner_output)
            self.assertIn('3.0: (unload l3 r1 c1)\n4.0: (move l3 l1 r1)\n\nPlan-Length:5\n', planner_output)
            self.assertIn('\nMetric (Search):5.0\nPlanning Time (msec):', planner_output)
            self.assertIn('\nHeuristic Time (msec): ', planner_output)
            self.assertIn('\nSearch Time (msec): ', planner_output)
            self.assertIn('\nExpanded Nodes:7\nStates Evaluated:8\n', planner_output)
            self.assertIn('\nFixed constraint violations during search (zero-crossing):0\nNumber of Dead-Ends detected:0\n', planner_output)
            self.assertIn('\nNumber of Duplicates detected:8\n', planner_output)

            for lm in final_report.log_messages:
                if lm.level == up.solvers.LogLevel.INFO:
                    self.assertEqual(planner_output, lm.message)
                else:
                    self.assertEqual(lm.level, up.solvers.LogLevel.ERROR)
                    self.assertEqual(lm.message, '')

    @skipIfSolverNotAvailable('opt-pddl-solver')
    def test_robot_loader_adv_with_short_timeout_and_output_stream(self):
        problem, right_plan = self.problems['robot_loader_adv'].problem, self.problems['robot_loader_adv'].plan
        output_stream = StringIO()
        with OneshotPlanner(name='opt-pddl-solver') as planner:
            self.assertNotEqual(planner, None)

            final_report = planner.solve(problem, timeout=VERYSMALL_TIMEOUT, output_stream=output_stream)
            plan = final_report.plan
            planner_output = output_stream.getvalue()
            self.assertTrue(plan is None or plan == right_plan)

            for lm in final_report.log_messages:
                if lm.level == up.solvers.LogLevel.INFO:
                    self.assertEqual(planner_output, lm.message)
                else:
                    self.assertEqual(lm.level, up.solvers.LogLevel.ERROR)
                    self.assertEqual(lm.message, '')



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
'''
This module defines the Parameter class. Both actions and fluents use this class to represent their parameters.
'''


import unified_planning as up


class Parameter:
    '''Represents an action parameter or a fluent parameter.
    A parameter has a name, and a type.'''
    def __init__(self, name: str, typename: 'up.model.types.Type'):
        self._name = name
        self._typename = typename

    def __repr__(self) -> str:
        return f'{str(self.type)} {self.name}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Parameter):
            return self._name == oth._name and self._typename == oth._typename
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + hash(self._typename)

    @property
    def name(self) -> str:
        '''Returns the parameter name.'''
        return self._name

    @property
    def type(self) -> 'up.model.types.Type':
        '''Returns the parameter type.'''
        return self._typename



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning.walkers as walkers

import unified_planning.environment
from unified_planning.walkers.identitydag import IdentityDagWalker
from unified_planning.model import FNode, OperatorKind
from unified_planning.exceptions import UPTypeError
from unified_planning.model import Expression
from typing import List, Dict


class Substituter(IdentityDagWalker):
    """Performs substitution into an expression """
    def __init__(self, env: 'unified_planning.environment.Environment'):
        IdentityDagWalker.__init__(self, env, True)
        self.env = env
        self.manager = env.expression_manager
        self.type_checker = env.type_checker

    def _get_key(self, expression, **kwargs):
        return expression

    def substitute(self, expression: FNode, substitutions: Dict[Expression, Expression] = {}) -> FNode:
        """Performs substitution into the given expression.

        Lets consider the examples:
        f = a & b
        subs = {a -> c, (c & b) -> d, (a & b) -> c}
        substitute(f, subs) = c

        f = a
        subs = {a -> c, c -> d}
        substitute(f, subs) = c

        f = a & b
        subs = {a -> 5, b -> c}
        substitute(f, subs) raises an UPTypeError

        Note that, since subs is a dictionary:
        f = a
        subs = {a -> b, a -> c}
        substitute(f, subs) = c
        """

        if len(substitutions) == 0:
            return expression
        new_substitutions: Dict[FNode, FNode] = {}
        for k, v in substitutions.items():
            new_k, new_v = self.manager.auto_promote(k, v)
            if self.type_checker.is_compatible_exp(new_v, new_k):
                new_substitutions[new_k] = new_v
            else:
                raise UPTypeError(
                    f"The expression type of {str(k)} is not compatible with the given substitution {str(v)}")
        return self.walk(expression, subs = new_substitutions)

    @walkers.handles(OperatorKind)
    def walk_replace_or_identity(self, expression: FNode, args: List[FNode], subs: Dict[FNode, FNode] = {}, **kwargs) -> FNode:
        res = subs.get(expression, None)
        if res is not None:
            return res
        else:
            return IdentityDagWalker.super(self, expression, args, **kwargs)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
import itertools
import tarski.fstrips # type: ignore
from fractions import Fraction
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.environment import Environment
from collections import OrderedDict
from typing import List, Optional, Union, Dict, cast
from tarski.syntax import Interval # type: ignore
from tarski.syntax.formulas import Formula, is_and, is_or, is_neg, is_atom # type: ignore
from tarski.syntax.formulas import Tautology, Contradiction, QuantifiedFormula, Quantifier # type: ignore
from tarski.syntax.terms import Term, CompoundTerm, BuiltinPredicateSymbol # type: ignore
from tarski.syntax.terms import Constant, Variable, BuiltinFunctionSymbol # type: ignore
from tarski.fstrips.fstrips import AddEffect, DelEffect, FunctionalEffect # type: ignore


def convert_tarski_formula(env: Environment, fluents: Dict[str, 'unified_planning.model.Fluent'],
                           objects: Dict[str, 'unified_planning.model.Object'],
                           action_parameters: Dict[str, 'unified_planning.model.Parameter'],
                           types: Dict[str, Optional['unified_planning.model.Type']],
                           formula: Union[Formula, Term]) -> 'unified_planning.model.FNode':
    """Converts a tarski formula in a unified_planning expression."""
    em = env.expression_manager
    if is_and(formula):
        children = [convert_tarski_formula(env, fluents, objects, action_parameters, types, f)
                    for f in formula.subformulas]
        return em.And(*children)
    elif is_or(formula):
        children = [convert_tarski_formula(env, fluents, objects, action_parameters, types, f)
                    for f in formula.subformulas]
        return em.Or(*children)
    elif is_neg(formula):
        assert len(formula.subformulas) == 1
        return em.Not(convert_tarski_formula(env, fluents, objects, action_parameters, types,
                                                  formula.subformulas[0]))
    elif is_atom(formula) or isinstance(formula, CompoundTerm):
        children = [convert_tarski_formula(env, fluents, objects, action_parameters, types, f)
                    for f in formula.subterms]
        if is_atom(formula):
            symbol = formula.predicate.symbol
        else:
            symbol = formula.symbol.name
        if symbol == BuiltinPredicateSymbol.EQ:
            assert len(children) == 2
            return em.Equals(children[0], children[1])
        elif symbol == BuiltinPredicateSymbol.NE:
            assert len(children) == 2
            return em.Not(em.Equals(children[0], children[1]))
        elif symbol == BuiltinPredicateSymbol.LT:
            assert len(children) == 2
            return em.LT(children[0], children[1])
        elif symbol == BuiltinPredicateSymbol.LE:
            assert len(children) == 2
            return em.LE(children[0], children[1])
        elif symbol == BuiltinPredicateSymbol.GT:
            assert len(children) == 2
            return em.GT(children[0], children[1])
        elif symbol == BuiltinPredicateSymbol.GE:
            assert len(children) == 2
            return em.GE(children[0], children[1])
        elif symbol == BuiltinFunctionSymbol.ADD:
            assert len(children) == 2
            return em.Plus(children[0], children[1])
        elif symbol == BuiltinFunctionSymbol.SUB:
            assert len(children) == 2
            return em.Minus(children[0], children[1])
        elif symbol == BuiltinFunctionSymbol.MUL:
            assert len(children) == 2
            return em.Times(children[0], children[1])
        elif symbol == BuiltinFunctionSymbol.DIV:
            assert len(children) == 2
            return em.Div(children[0], children[1])
        elif symbol in fluents:
            return fluents[symbol](*children)
        else:
            raise UPProblemDefinitionError(symbol + ' not supported!')
    elif isinstance(formula, Constant):
        if formula.sort.name == 'number':
            return em.Real(Fraction(float(formula.name)))
        elif isinstance(formula.sort, tarski.syntax.Interval):
            if formula.sort.language.is_subtype(\
                formula.sort, formula.sort.language.Integer)\
                or formula.sort.language.is_subtype(\
                    formula.sort, formula.sort.language.Natural):
                return em.Int(int(formula.name))
            elif formula.sort.language.is_subtype(\
                formula.sort, formula.sort.language.Real):
                return em.Real(Fraction(float(formula.name)))
            else:
                raise NotImplementedError
        elif formula.name in objects:
            return em.ObjectExp(objects[formula.name])
        else:
            raise UPProblemDefinitionError(formula + ' not supported!')
    elif isinstance(formula, Variable):
        if formula.symbol in action_parameters:
            return em.ParameterExp(action_parameters[formula.symbol])
        else:
            return em.VariableExp(unified_planning.model.Variable(formula.symbol, \
                cast(unified_planning.model.Type, _convert_type_and_update_dict(formula.sort, types, env.type_manager, formula.sort.language))))
    elif isinstance(formula, QuantifiedFormula):
        expression = convert_tarski_formula(env, fluents, objects, action_parameters, types, formula.formula)
        variables = [unified_planning.model.Variable(v.symbol,
            cast(unified_planning.model.Type, _convert_type_and_update_dict(v.sort, types, env.type_manager, v.sort.language)))
            for v in formula.variables]
        if formula.quantifier == Quantifier.Exists:
            return em.Exists(expression, *variables)
        elif formula.quantifier == Quantifier.Forall:
            return em.Forall(expression, *variables)
        else:
            raise NotImplementedError
    elif isinstance(formula, Tautology):
        return em.TRUE()
    elif isinstance(formula, Contradiction):
        return em.FALSE()
    else:
        raise UPProblemDefinitionError(str(formula) + ' not supported!')


def _check_if_tarski_problem_uses_object_type(tarski_problem: tarski.fstrips.Problem) -> bool:
    lang = tarski_problem.language
    for p in lang.predicates:
        if str(p.name) in ['=', '!=', '<', '<=', '>', '>=']:
            continue
        for t in p.sort:
            if str(t.name) == 'object':
                return True
    for p in lang.functions:
        if str(p.name) in ['ite', '@', '+', '-', '*', '/', '**', '%', 'sqrt']:
            continue
        for t in p.domain:
            if str(t.name) == 'object':
                return True
        func_sort = p.sort[-1]
        if str(func_sort.name) == 'object':
            return True
    for c in lang.constants():
        if str(c.sort.name) == 'object':
            return True
    for a_name in tarski_problem.actions:
        a = tarski_problem.get_action(a_name)
        for p in a.parameters:
            if str(p.sort.name) == 'object':
                return True
    return False


def _convert_type_and_update_dict(sort: tarski.syntax.Sort, types_dict: Dict[str, Optional['unified_planning.model.Type']], \
                tm: 'unified_planning.model.TypeManager', lang: 'tarski.fol.FirstOrderLanguage') \
                -> Optional['unified_planning.model.Type']:
    '''Converts a tarski type in a unified_planning type and inserts it into the types_dict.
        Important NOTE: This function modifies the parameter types_dict.'''
    if str(sort.name) in types_dict: # type already defined
        return types_dict[str(sort.name)]
    if isinstance(sort, Interval): # if the type is an Interval
        if sort == lang.Integer:
            up_type = tm.IntType()
        elif sort == lang.Natural:
            up_type = tm.IntType(lower_bound=0)
        elif sort == lang.Real:
            up_type = tm.RealType()
        elif sort.encode == lang.Integer.encode:
            up_type = tm.IntType(sort.lower_bound, sort.upper_bound)
        elif sort.encode == lang.Natural.encode:
            up_type = tm.IntType(0, sort.upper_bound)
        elif sort.encode == lang.Real.encode:
            up_type = tm.RealType(sort.lower_bound, sort.upper_bound)
        else:
            raise NotImplementedError
    else: # the type should be a user_type
        tarski_father = tarski.syntax.sorts.parent(sort)
        up_father: Optional['unified_planning.model.Type'] = None
        if tarski_father is not None:
            up_father = _convert_type_and_update_dict(tarski_father, types_dict, tm, lang)
        up_type = tm.UserType(str(sort.name), up_father)
    assert up_type is not None #sanity check
    types_dict[str(sort.name)] = up_type
    return up_type


def convert_problem_from_tarski(env: Environment, tarski_problem: tarski.fstrips.Problem) -> 'unified_planning.model.Problem':
    """Converts a tarski problem in a unified_planning.Problem."""
    em = env.expression_manager
    tm = env.type_manager
    lang = tarski_problem.language
    problem = unified_planning.model.Problem(tarski_problem.name)

    # Convert types
    types: Dict[str, Optional['unified_planning.model.Type']] = {}
    uses_object_type: bool = _check_if_tarski_problem_uses_object_type(tarski_problem)
    if not uses_object_type:
        types['object'] = None # we set object as None, so when it is the father of a type in tarski, in UP it will be None.
    for t in lang.sorts:
        #types will be filled with the needed types in this loop.
        _convert_type_and_update_dict(t, types, tm, lang)


    # Convert predicates and functions
    fluents = {}
    for p in lang.predicates:
        if str(p.name) in ['=', '!=', '<', '<=', '>', '>=']:
            continue
        signature: OrderedDict[str, 'unified_planning.model.Type'] = OrderedDict()
        for i, t in enumerate(p.sort):
            type = types[str(t.name)]
            assert type is not None
            signature[f'p{str(i+1)}'] = type
        fluent = unified_planning.model.Fluent(p.name, tm.BoolType(), signature)
        fluents[fluent.name] = fluent
        problem.add_fluent(fluent)
    for p in lang.functions:
        if str(p.name) in ['ite', '@', '+', '-', '*', '/', '**', '%', 'sqrt']:
            continue
        signature = OrderedDict()
        for i, t in enumerate(p.domain):
            type = types[str(t.name)]
            assert type is not None
            signature[f'p{str(i+1)}'] = type
        func_sort = p.sort[-1]
        if isinstance(func_sort, Interval):
            if func_sort.encode == lang.Real.encode:
                if func_sort.name == 'Real' or func_sort.name == 'number':
                    fluent = unified_planning.model.Fluent(p.name, tm.RealType(), signature)
                else:
                    fluent = unified_planning.model.Fluent(p.name, tm.RealType(lower_bound=\
                        Fraction(func_sort.lower_bound), upper_bound=Fraction(func_sort.upper_bound)), signature)
            else:
                assert func_sort.encode == lang.Integer.encode or func_sort.encode == lang.Natural.encode
                if func_sort.name == 'Integer':
                    fluent = unified_planning.model.Fluent(p.name, tm.IntType(), signature)
                elif func_sort.name == 'Natual':
                    fluent = unified_planning.model.Fluent(p.name, tm.IntType(lower_bound=0), signature)
                else:
                    fluent = unified_planning.model.Fluent(p.name, tm.IntType(lower_bound=\
                        func_sort.lower_bound, upper_bound=func_sort.upper_bound), signature)
        else:
            fluent = unified_planning.model.Fluent(p.name, types[func_sort.name], signature)
        fluents[fluent.name] = fluent
        problem.add_fluent(fluent)

    # Convert objects
    objects = {}
    for c in lang.constants():
        type = types[str(c.sort.name)]
        assert type is not None
        o = unified_planning.model.Object(str(c.name), type)
        objects[o.name] = o
        problem.add_object(o)

    # Convert actions
    for a_name in tarski_problem.actions:
        a = tarski_problem.get_action(a_name)
        parameters: OrderedDict[str, 'unified_planning.model.Type'] = OrderedDict()
        for p in a.parameters:
            type = types[str(p.sort.name)]
            assert type is not None
            parameters[p.symbol] = type
        action = unified_planning.model.InstantaneousAction(a_name, parameters)
        action_parameters = {}
        for p in parameters.keys():
            action_parameters[p] = action.parameter(p)
        f = convert_tarski_formula(env, fluents, objects, action_parameters, types, a.precondition)
        action.add_precondition(f)
        for eff in a.effects:
            condition = convert_tarski_formula(env, fluents, objects, action_parameters, types, eff.condition)
            if isinstance(eff, AddEffect):
                f = convert_tarski_formula(env, fluents, objects, action_parameters, types, eff.atom)
                action.add_effect(f, True, condition)
            elif isinstance(eff, DelEffect):
                f = convert_tarski_formula(env, fluents, objects, action_parameters, types, eff.atom)
                action.add_effect(f, False, condition)
            elif isinstance(eff, FunctionalEffect):
                lhs = convert_tarski_formula(env, fluents, objects, action_parameters, types, eff.lhs)
                rhs = convert_tarski_formula(env, fluents, objects, action_parameters, types, eff.rhs)
                action.add_effect(lhs, rhs, condition)
            else:
                raise UPProblemDefinitionError(eff + ' not supported!')
        problem.add_action(action)

    # Set initial values
    initial_values = {}
    for fluent in fluents.values():
        l = [problem.objects_hierarchy(p.type) for p in fluent.signature]
        if fluent.type.is_bool_type():
            default_value = em.FALSE()
        elif fluent.type.is_real_type():
            default_value = em.Real(Fraction(0))
        elif fluent.type.is_int_type():
            default_value = em.Int(0)
        elif fluent.type.is_user_type():
            continue
        if len(l) == 0:
            initial_values[em.FluentExp(fluent)] = default_value
        else:
            for args in itertools.product(*l):
                initial_values[fluent(*args)] = default_value
    for i in tarski_problem.init.as_atoms():
        if isinstance(i, tuple):
            lhs = convert_tarski_formula(env, fluents, objects, {}, types, i[0])
            rhs = convert_tarski_formula(env, fluents, objects, {}, types, i[1])
            initial_values[lhs] = rhs
        else:
            f = convert_tarski_formula(env, fluents, objects, {}, types, i)
            initial_values[f] = em.TRUE()
    for lhs, rhs in initial_values.items():
        problem.set_initial_value(lhs, rhs)

    # Convert goals
    problem.add_goal(convert_tarski_formula(env, fluents, objects, {}, types, tarski_problem.goal))

    return problem



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
import sys

from decimal import Decimal, localcontext
from warnings import warn

import unified_planning as up
import unified_planning.environment
import unified_planning.walkers as walkers
from unified_planning.model import DurativeAction
from unified_planning.exceptions import UPTypeError, UPProblemDefinitionError
from typing import IO, List, Optional
from io import StringIO
from functools import reduce


class ConverterToPDDLString(walkers.DagWalker):
    '''Expression converter to a PDDL string.'''

    DECIMAL_PRECISION = 10 # Number of decimal places to print real constants

    def __init__(self, env: 'up.environment.Environment'):
        walkers.DagWalker.__init__(self)
        self.simplifier = walkers.Simplifier(env)

    def convert(self, expression):
        '''Converts the given expression to a PDDL string.'''
        return self.walk(self.simplifier.simplify(expression))

    def walk_exists(self, expression, args):
        assert len(args) == 1
        vars_string_list = [f"?{v.name} - {str(v.type)}" for v in expression.variables()]
        return f'(exists ({" ".join(vars_string_list)})\n {args[0]})'

    def walk_forall(self, expression, args):
        assert len(args) == 1
        vars_string_list = [f"?{v.name} - {str(v.type)}" for v in expression.variables()]
        return f'(forall ({" ".join(vars_string_list)})\n {args[0]})'

    def walk_variable_exp(self, expression, args):
        assert len(args) == 0
        return f'?{expression.variable().name}'

    def walk_and(self, expression, args):
        assert len(args) > 1
        return f'(and {" ".join(args)})'

    def walk_or(self, expression, args):
        assert len(args) > 1
        return f'(or {" ".join(args)})'

    def walk_not(self, expression, args):
        assert len(args) == 1
        return f'(not {args[0]})'

    def walk_implies(self, expression, args):
        assert len(args) == 2
        return f'(imply {args[0]} {args[1]})'

    def walk_iff(self, expression, args):
        assert len(args) == 2
        return f'(and (imply {args[0]} {args[1]}) (imply {args[1]} {args[0]}) )'

    def walk_fluent_exp(self, expression, args):
        fluent = expression.fluent()
        return f'({fluent.name}{" " if len(args) > 0 else ""}{" ".join(args)})'

    def walk_param_exp(self, expression, args):
        assert len(args) == 0
        p = expression.parameter()
        return f'?{p.name}'

    def walk_object_exp(self, expression, args):
        assert len(args) == 0
        o = expression.object()
        return f'{o.name}'

    def walk_bool_constant(self, expression, args):
        raise up.exceptions.UPUnreachableCodeError

    def walk_real_constant(self, expression, args):
        assert len(args) == 0
        frac = expression.constant_value()

        with localcontext() as ctx:
            ctx.prec = self.DECIMAL_PRECISION
            dec = frac.numerator / Decimal(frac.denominator, ctx)

            if Fraction(dec) != frac:
                warn("The PDDL printer cannot exactly represent the real constant '%s'" % frac)
            return str(dec)

    def walk_int_constant(self, expression, args):
        assert len(args) == 0
        return str(expression.constant_value())

    def walk_plus(self, expression, args):
        assert len(args) > 1
        return reduce(lambda x, y: f'(+ {y} {x})', args)

    def walk_minus(self, expression, args):
        assert len(args) == 2
        return f'(- {args[0]} {args[1]})'

    def walk_times(self, expression, args):
        assert len(args) > 1
        return reduce(lambda x, y: f'(* {y} {x})', args)

    def walk_div(self, expression, args):
        assert len(args) == 2
        return f'(/ {args[0]} {args[1]})'

    def walk_le(self, expression, args):
        assert len(args) == 2
        return f'(<= {args[0]} {args[1]})'

    def walk_lt(self, expression, args):
        assert len(args) == 2
        return f'(< {args[0]} {args[1]})'

    def walk_equals(self, expression, args):
        assert len(args) == 2
        return f'(= {args[0]} {args[1]})'


class PDDLWriter:
    '''This class can be used to write a Problem in PDDL.'''

    def __init__(self, problem: 'up.model.Problem', needs_requirements: bool = True):
        self.problem = problem
        self.needs_requirements = needs_requirements
        self.object_freshname = 'object'

    def _type_name_or_object_freshname(self, type: 'unified_planning.model.Type') -> str:
        return type.name if type.name != "object" else self.object_freshname # type: ignore

    def _write_domain(self, out: IO[str]):
        problem_kind = self.problem.kind
        if problem_kind.has_intermediate_conditions_and_effects(): # type: ignore
            raise UPProblemDefinitionError('PDDL2.1 does not support ICE.\nICE are Intermediate Conditions and Effects therefore when an Effect (or Condition) are not at StartTIming(0) or EndTIming(0).')
        if problem_kind.has_timed_effect() or problem_kind.has_timed_goals(): # type: ignore
            raise UPProblemDefinitionError('PDDL2.1 does not support timed effects or timed goals.')
        out.write('(define ')
        if self.problem.name is None:
            name = 'pddl'
        else:
            name = f'{self.problem.name}'
        out.write(f'(domain {name}-domain)\n')

        if self.needs_requirements:
            out.write(' (:requirements :strips')
            if problem_kind.has_flat_typing(): # type: ignore
                out.write(' :typing')
            if problem_kind.has_negative_conditions(): # type: ignore
                out.write(' :negative-preconditions')
            if problem_kind.has_disjunctive_conditions(): # type: ignore
                out.write(' :disjunctive-preconditions')
            if problem_kind.has_equality(): # type: ignore
                out.write(' :equality')
            if (problem_kind.has_continuous_numbers() or # type: ignore
                problem_kind.has_discrete_numbers()): # type: ignore
                out.write(' :numeric-fluents')
            if problem_kind.has_conditional_effects(): # type: ignore
                out.write(' :conditional-effects')
            if problem_kind.has_existential_conditions(): # type: ignore
                out.write(' :existential-preconditions')
            if problem_kind.has_universal_conditions(): # type: ignore
                out.write(' :universal-preconditions')
            if (problem_kind.has_continuous_time() or # type: ignore
                problem_kind.has_discrete_time()): # type: ignore
                out.write(' :durative-actions')
            if problem_kind.has_duration_inequalities(): # type: ignore
                out.write(' :duration-inequalities')
            if (self.problem.kind.has_actions_cost() or # type: ignore
                self.problem.kind.has_plan_length()): # type: ignore
                out.write(' :action-costs')
            out.write(')\n')


        if problem_kind.has_hierarchical_typing(): # type: ignore
            while self.problem.has_type(self.object_freshname):
                self.object_freshname = self.object_freshname + '_'
            user_types_hierarchy = self.problem.user_types_hierarchy
            out.write(f' (:types\n')
            stack: List['unified_planning.model.Type'] = user_types_hierarchy[None] if None in user_types_hierarchy else []
            out.write(f'    {" ".join(self._type_name_or_object_freshname(t) for t in stack)} - object\n')
            while stack:
                current_type = stack.pop()
                direct_sons: List['unified_planning.model.Type'] = user_types_hierarchy[current_type]
                if direct_sons:
                    stack.extend(direct_sons)
                    out.write(f'    {" ".join([self._type_name_or_object_freshname(t) for t in direct_sons])} - {self._type_name_or_object_freshname(current_type)}\n')
            out.write(' )\n')
        else:
            out.write(f' (:types {" ".join([t.name for t in self.problem.user_types])})\n' if len(self.problem.user_types) > 0 else '') # type: ignore

        predicates = []
        functions = []
        for f in self.problem.fluents:
            if f.type.is_bool_type():
                params = []
                i = 0
                for param in f.signature:
                    if param.type.is_user_type():
                        params.append(f' ?{param.name} - {self._type_name_or_object_freshname(param.type)}')
                        i += 1
                    else:
                        raise UPTypeError('PDDL supports only user type parameters')
                predicates.append(f'({f.name}{"".join(params)})')
            elif f.type.is_int_type() or f.type.is_real_type():
                params = []
                i = 0
                for param in f.signature:
                    if param.type.is_user_type():
                        params.append(f' ?{param.name} - {self._type_name_or_object_freshname(param.type)}')
                        i += 1
                    else:
                        raise UPTypeError('PDDL supports only user type parameters')
                functions.append(f'({f.name}{"".join(params)})')
            else:
                raise UPTypeError('PDDL supports only boolean and numerical fluents')
        if (self.problem.kind.has_actions_cost() or # type: ignore
            self.problem.kind.has_plan_length()): # type: ignore
            functions.append('(total-cost)')
        out.write(f' (:predicates {" ".join(predicates)})\n' if len(predicates) > 0 else '')
        out.write(f' (:functions {" ".join(functions)})\n' if len(functions) > 0 else '')

        converter = ConverterToPDDLString(self.problem.env)
        costs = {}
        metrics = self.problem.quality_metrics
        if len(metrics) == 1:
            metric = metrics[0]
            if isinstance(metric, up.model.metrics.MinimizeActionCosts):
                for a in self.problem.actions:
                    costs[a] = metric.get_action_cost(a)
            elif isinstance(metric, up.model.metrics.MinimizeSequentialPlanLength):
                for a in self.problem.actions:
                    costs[a] = self.problem.env.expression_manager.Int(1)
        elif len(metrics) > 1:
            raise up.exceptions.UPUnsupportedProblemTypeError('Only one metric is supported!')
        for a in self.problem.actions:
            if isinstance(a, up.model.InstantaneousAction):
                out.write(f' (:action {a.name}')
                out.write(f'\n  :parameters (')
                for ap in a.parameters:
                    if ap.type.is_user_type():
                        out.write(f' ?{ap.name} - {self._type_name_or_object_freshname(ap.type)}')
                    else:
                        raise UPTypeError('PDDL supports only user type parameters')
                out.write(')')
                if len(a.preconditions) > 0:
                    out.write(f'\n  :precondition (and {" ".join([converter.convert(p) for p in a.preconditions])})')
                if len(a.effects) > 0:
                    out.write('\n  :effect (and')
                    for e in a.effects:
                        if e.is_conditional():
                            out.write(f' (when {converter.convert(e.condition)}')
                        if e.value.is_true():
                            out.write(f' {converter.convert(e.fluent)}')
                        elif e.value.is_false():
                            out.write(f' (not {converter.convert(e.fluent)})')
                        elif e.is_increase():
                            out.write(f' (increase {converter.convert(e.fluent)} {converter.convert(e.value)})')
                        elif e.is_decrease():
                            out.write(f' (decrease {converter.convert(e.fluent)} {converter.convert(e.value)})')
                        else:
                            out.write(f' (assign {converter.convert(e.fluent)} {converter.convert(e.value)})')
                        if e.is_conditional():
                            out.write(f')')

                    if a in costs:
                        out.write(f' (increase total-cost {converter.convert(costs[a])})')
                    out.write(')')
                out.write(')\n')
            elif isinstance(a, DurativeAction):
                out.write(f' (:durative-action {a.name}')
                out.write(f'\n  :parameters (')
                for ap in a.parameters:
                    if ap.type.is_user_type():
                        out.write(f' ?{ap.name} - {self._type_name_or_object_freshname(ap.type)}')
                    else:
                        raise UPTypeError('PDDL supports only user type parameters')
                out.write(')')
                l, r = a.duration.lower, a.duration.upper
                if l == r:
                    out.write(f'\n  :duration (= ?duration {str(l)})')
                else:
                    out.write(f'\n  :duration (and ')
                    if a.duration.is_left_open():
                        out.write(f'(> ?duration {str(l)})')
                    else:
                        out.write(f'(>= ?duration {str(l)})')
                    if a.duration.is_right_open():
                        out.write(f'(< ?duration {str(r)})')
                    else:
                        out.write(f'(<= ?duration {str(r)})')
                    out.write(')')
                if len(a.conditions) > 0:
                    out.write(f'\n  :condition (and ')
                    for interval, cl in a.conditions.items():
                        for c in cl:
                            if interval.lower == interval.upper:
                                if interval.lower.is_from_start():
                                    out.write(f'(at start {converter.convert(c)})')
                                else:
                                    out.write(f'(at end {converter.convert(c)})')
                            else:
                                if not interval.is_left_open():
                                    out.write(f'(at start {converter.convert(c)})')
                                out.write(f'(over all {converter.convert(c)})')
                                if not interval.is_right_open():
                                    out.write(f'(at end {converter.convert(c)})')
                    out.write(')')
                if len(a.effects) > 0:
                    out.write('\n  :effect (and')
                    for t, el in a.effects.items():
                        for e in el:
                            if t.is_from_start():
                                out.write(f' (at start')
                            else:
                                out.write(f' (at end')
                            if e.is_conditional():
                                out.write(f' (when {converter.convert(e.condition)}')
                            if e.value.is_true():
                                out.write(f' {converter.convert(e.fluent)}')
                            elif e.value.is_false():
                                out.write(f' (not {converter.convert(e.fluent)})')
                            elif e.is_increase():
                                out.write(f' (increase {converter.convert(e.fluent)} {converter.convert(e.value)})')
                            elif e.is_decrease():
                                out.write(f' (decrease {converter.convert(e.fluent)} {converter.convert(e.value)})')
                            else:
                                out.write(f' (assign {converter.convert(e.fluent)} {converter.convert(e.value)})')
                            if e.is_conditional():
                                out.write(f')')
                            out.write(')')
                    if a in costs:
                        out.write(f' (at end (increase total-cost {converter.convert(costs[a])}))')
                    out.write(')')
                out.write(')\n')
            else:
                raise NotImplementedError
        out.write(')\n')

    def _write_problem(self, out: IO[str]):
        if self.problem.name is None:
            name = 'pddl'
        else:
            name = f'{self.problem.name}'
        out.write(f'(define (problem {name}-problem)\n')
        out.write(f' (:domain {name}-domain)\n')
        if len(self.problem.user_types) > 0:
            out.write(' (:objects ')
            for t in self.problem.user_types:
                objects: List['unified_planning.model.Object'] = list(self.problem.objects(t))
                if len(objects) > 0:
                    out.write(f'\n   {" ".join([o.name for o in objects])} - {self._type_name_or_object_freshname(t)}')
            out.write('\n )\n')
        converter = ConverterToPDDLString(self.problem.env)
        out.write(' (:init')
        for f, v in self.problem.initial_values.items():
            if v.is_true():
                out.write(f' {converter.convert(f)}')
            elif v.is_false():
                pass
            else:
                out.write(f' (= {converter.convert(f)} {converter.convert(v)})')
        if self.problem.kind.has_actions_cost(): # type: ignore
            out.write(f' (= total-cost 0)')
        out.write(')\n')
        out.write(f' (:goal (and {" ".join([converter.convert(p) for p in self.problem.goals])}))\n')
        metrics = self.problem.quality_metrics
        if len(metrics) == 1:
            metric = metrics[0]
            out.write(' (:metric ')
            if isinstance(metric, up.model.metrics.MinimizeExpressionOnFinalState):
                out.write(f'minimize {metric.expression}')
            elif isinstance(metric, up.model.metrics.MaximizeExpressionOnFinalState):
                out.write(f'maximize {metric.expression}')
            elif (isinstance(metric, up.model.metrics.MinimizeActionCosts) or
                  isinstance(metric, up.model.metrics.MinimizeSequentialPlanLength)):
                out.write(f'minimize total-cost')
            elif isinstance(metric, up.model.metrics.MinimizeMakespan):
                out.write(f'minimize total-time')
            else:
                raise NotImplementedError
            out.write(')\n')
        elif len(metrics) > 1:
            raise up.exceptions.UPUnsupportedProblemTypeError('Only one metric is supported!')
        out.write(')\n')

    def print_domain(self):
        '''Prints to std output the PDDL domain.'''
        self._write_domain(sys.stdout)

    def print_problem(self):
        '''Prints to std output the PDDL problem.'''
        self._write_problem(sys.stdout)

    def get_domain(self) -> str:
        '''Returns the PDDL domain.'''
        out = StringIO()
        self._write_domain(out)
        return out.getvalue()

    def get_problem(self) -> str:
        '''Returns the PDDL problem.'''
        out = StringIO()
        self._write_problem(out)
        return out.getvalue()

    def write_domain(self, filename: str):
        '''Dumps to file the PDDL domain.'''
        with open(filename, 'w') as f:
            self._write_domain(f)

    def write_problem(self, filename: str):
        '''Dumps to file the PDDL problem.'''
        with open(filename, 'w') as f:
            self._write_problem(f)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from collections import namedtuple

Example = namedtuple('Example', ['problem', 'plan'])

def get_example_problems():
    problems = {}

    # robot_real_constants
    #this version of the problem robot has reals instead of integers as constants
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    battery_charge = Fluent('battery_charge', RealType(0, 100))
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge, 10.0))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10.0))
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_real_constants')
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100.0)
    problem.add_goal(robot_at(l2))
    plan = unified_planning.plan.SequentialPlan([unified_planning.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))])
    robot = Example(problem=problem, plan=plan)
    problems['robot_real_constants'] = robot

    # robot_int_battery
    #this version of the problem robot has the battery charge fluent represented as an int instead of a real
    Location = UserType('Location')
    robot_at = Fluent('robot_at', BoolType(), position=Location)
    battery_charge = Fluent('battery_charge', IntType(0, 100))
    move = InstantaneousAction('move', l_from=Location, l_to=Location)
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10))
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_int_battery')
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = unified_planning.plan.SequentialPlan([unified_planning.plan.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))])
    robot = Example(problem=problem, plan=plan)
    problems['robot_int_battery'] = robot

    #robot fluent of user_type with int ID
    Int_t = IntType(0,1)
    Location = UserType('Location')
    is_at = Fluent('is_at', Location, id=Int_t)
    move = InstantaneousAction('move', robot=Int_t, l_from=Location, l_to=Location)
    robot = move.parameter('robot')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(Equals(is_at(robot), l_from))
    move.add_precondition(Not(Equals(is_at(robot), l_to)))
    move.add_effect(is_at(robot), l_to)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    problem = Problem('robot_fluent_of_user_type_with_int_id')
    problem.add_fluent(is_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(is_at(Int(0)), l1)
    problem.set_initial_value(is_at(1), l1)
    plan = unified_planning.plan.SequentialPlan([unified_planning.plan.ActionInstance(move, (Int(0), ObjectExp(l1), ObjectExp(l2))),
                                    unified_planning.plan.ActionInstance(move, (Int(1), ObjectExp(l1), ObjectExp(l2)))])
    robot_fluent_of_user_type_with_int_id = Example(problem=problem, plan=plan)
    problems['robot_fluent_of_user_type_with_int_id'] = robot_fluent_of_user_type_with_int_id

    #robot locations connected without battery
    Location = UserType('Location')
    Robot = UserType('Robot')
    is_at = Fluent('is_at', BoolType(), position=Location, robot=Robot)
    is_connected = Fluent('is_connected', BoolType(), location_1=Location, location_2=Location)
    move = InstantaneousAction('move', robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter('robot')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move_2 = InstantaneousAction('move_2', robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter('robot')
    l_from = move_2.parameter('l_from')
    l_to = move_2.parameter('l_to')
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable('mid_loc', Location)
    #(E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(Exists(And(Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                            Or(is_connected(l_from, mid_location), is_connected(mid_location, l_from)),
                            Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to))), mid_location))
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    l1 = Object('l1', Location)
    l2 = Object('l2', Location)
    l3 = Object('l3', Location)
    l4 = Object('l4', Location)
    l5 = Object('l5', Location)
    r1 = Object('r1', Robot)
    problem = Problem('robot_locations_connected_without_battery')
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.add_goal(is_at(l5, r1))
    plan = unified_planning.plan.SequentialPlan([unified_planning.plan.ActionInstance(move_2, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l3))),
                                unified_planning.plan.ActionInstance(move_2, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l5)))])
    robot_locations_connected_without_battery = Example(problem=problem, plan=plan)
    problems['robot_locations_connected_without_battery'] = robot_locations_connected_without_battery

    # hierarchical blocks world exists
    Entity = UserType('Entity', None) # None can be avoided
    Location = UserType('Location', Entity)
    Unmovable = UserType('Unmovable', Location)
    TableSpace = UserType('TableSpace', Unmovable)
    Movable = UserType('Movable', Location)
    Block = UserType('Block', Movable)
    clear = Fluent('clear', BoolType(), space=Location)
    on = Fluent('on', BoolType(), object=Movable, space=Location)

    move = InstantaneousAction('move', item=Movable, l_from=Location, l_to=Location)
    item = move.parameter('item')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem('hierarchical_blocks_world_exists')
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object('ts_1', TableSpace)
    ts_2 = Object('ts_2', TableSpace)
    ts_3 = Object('ts_3', TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object('block_1', Block)
    block_2 = Object('block_2', Block)
    block_3 = Object('block_3', Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    m_var = Variable('m_var', Movable) 
    problem.add_goal(Exists(on(m_var, block_1), m_var))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plan.SequentialPlan([
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))), 
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2)))])
    hierarchical_blocks_world_exists = Example(problem=problem, plan=plan)
    problems['hierarchical_blocks_world_exists'] = hierarchical_blocks_world_exists

    # hierarchical blocks world object as root
    object = UserType('object')
    Entity = UserType('Entity', object)
    Location = UserType('Location', Entity)
    Unmovable = UserType('Unmovable', Location)
    TableSpace = UserType('TableSpace', Unmovable)
    Movable = UserType('Movable', Location)
    Block = UserType('Block', Movable)
    clear = Fluent('clear', BoolType(), space=Location)
    on = Fluent('on', BoolType(), object=Movable, space=Location)

    move = InstantaneousAction('move', item=Movable, l_from=Location, l_to=Location)
    item = move.parameter('item')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem('hierarchical_blocks_world_object_as_root')
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object('ts_1', TableSpace)
    ts_2 = Object('ts_2', TableSpace)
    ts_3 = Object('ts_3', TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object('block_1', Block)
    block_2 = Object('block_2', Block)
    block_3 = Object('block_3', Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plan.SequentialPlan([
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))), 
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2)))])
    hierarchical_blocks_world_object_as_root = Example(problem=problem, plan=plan)
    problems['hierarchical_blocks_world_object_as_root'] = hierarchical_blocks_world_object_as_root

    # hierarchical blocks world with object
    Entity = UserType('Entity', None) # None can be avoided
    object = UserType('object', Entity)
    Unmovable = UserType('Unmovable', object)
    TableSpace = UserType('TableSpace', Unmovable)
    Movable = UserType('Movable', object)
    Block = UserType('Block', Movable)
    clear = Fluent('clear', BoolType(), space=object)
    on = Fluent('on', BoolType(), object=Movable, space=object)

    move = InstantaneousAction('move', item=Movable, l_from=object, l_to=object)
    item = move.parameter('item')
    l_from = move.parameter('l_from')
    l_to = move.parameter('l_to')
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem('hierarchical_blocks_world_with_object')
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object('ts_1', TableSpace)
    ts_2 = Object('ts_2', TableSpace)
    ts_3 = Object('ts_3', TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object('block_1', Block)
    block_2 = Object('block_2', Block)
    block_3 = Object('block_3', Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plan.SequentialPlan([
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))), 
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))),
            unified_planning.plan.ActionInstance(move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2)))])
    hierarchical_blocks_world_with_object = Example(problem=problem, plan=plan)
    problems['hierarchical_blocks_world_with_object'] = hierarchical_blocks_world_with_object

    return problems
